/*
* Copyright (C) 2009, Oy L M Ericsson Ab, NomadicLab <pubsub@nomadiclab.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* Alternatively, this software may be distributed under the terms of the BSD
* license.
*
* See LICENSE and COPYING for more details.
*/ 

%module libpsirp_rb
%{
#include <sys/param.h>
#include "../include/libpsirp.h"
%}

typedef unsigned long u_int64_t;
typedef void * psirp_pub_t;
typedef void * caddr_t;

%typemap(in) psirp_id_t * {
    struct RString *rs;
    int id_len;

    rs = RSTRING($input);
    
    $1 = ($1_ltype)malloc(sizeof($*1_ltype));
    id_len = sizeof(*$1);
    memset($1, 0, id_len);
    memcpy($1, rs->ptr, (rs->len < id_len) ? rs->len : id_len);
}
%typemap(freearg) psirp_id_t *, psirp_id_t * {
    if ($1) {
        free($1);
        $1 = NULL;
    }
}

%typemap(out) psirp_id_t * {
    VALUE v = rb_str_new((char *)$1, sizeof(*$1));
    $result = SWIG_Ruby_AppendOutput($result, v);
}

%typemap(in, numinputs=0) psirp_pub_t * {
    $1 = ($1_ltype)malloc(sizeof($*1_ltype));
    memset($1, 0, sizeof($*1_ltype));
}
%typemap(argout) psirp_pub_t * {
    VALUE v = SWIG_NewPointerObj(SWIG_as_voidptr(*$1),
                                 $descriptor(psirp_pub_t),
                                 0);
    $result = SWIG_Ruby_AppendOutput($result, v);
}
%typemap(freearg) psirp_pub_t * {
    if ($1) {
        free($1);
        $1 = NULL;
    }
}

%inline %{
VALUE psirp_rb_data_to_buffer(caddr_t data, u_int64_t len) {
    /* XXX */

    NEWOBJ(str, struct RString);
    OBJSETUP(str, rb_cString, T_STRING);

    str->ptr = data;
    str->len = len;
    str->aux.capa = len;
    
    OBJ_TAINT(str);
    OBJ_FREEZE(str);
    
    return (VALUE)str;
}

VALUE psirp_rb_buffer(psirp_pub_t pub) {
    return psirp_rb_data_to_buffer(psirp_pub_data(pub),
                                   psirp_pub_data_len(pub));
}

psirp_pub_t psirp_rb_publish_string(psirp_id_t *sid, psirp_id_t *rid,
                                    VALUE str) {
    /* XXX */

    struct RString *rs;
    psirp_pub_t pub;

    rs = RSTRING(str);

    psirp_create(rs->len, &pub);
    memcpy(psirp_pub_data(pub), rs->ptr, rs->len); 
    psirp_publish(sid, rid, pub);

    return pub;
}
%}

%constant int PSIRP_ID_LEN   = PSIRP_ID_LEN;
%constant int NOTE_PUBLISH   = NOTE_PUBLISH;
%constant int NOTE_SUBSCRIBE = NOTE_SUBSCRIBE;

%include "../include/libpsirp.h"
